import sys

import requests
import urllib3

from sdkite.http.exceptions import HTTPConnectionError, HTTPError, HTTPTimeoutError
from sdkite.http.model import HTTPHeaderDict, HTTPRequest, HTTPResponse
from sdkite.utils import walk_exception_context

if sys.version_info < (3, 8):  # pragma: no cover
    from backports.cached_property import cached_property
else:  # pragma: no cover
    from functools import cached_property

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterator
else:  # pragma: no cover
    from collections.abc import Iterator


class HTTPResponseRequests(HTTPResponse):
    def __init__(self, response: requests.Response) -> None:
        self._response = response

    @property
    def raw(self) -> requests.Response:
        return self._response

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def reason(self) -> str:
        return self._response.reason

    @cached_property
    def headers(self) -> HTTPHeaderDict:
        return HTTPHeaderDict(self._response.raw.headers)

    @cached_property
    def data_stream(self) -> Iterator[bytes]:
        return self._response.iter_content()

    @property
    def data_bytes(self) -> bytes:
        return self._response.content

    @property
    def data_str(self) -> str:
        return self._response.text

    @cached_property
    def data_json(self) -> object:
        return self._response.json()

    def _close(self) -> None:
        self._response.close()


def _extract_exception(exception: BaseException) -> BaseException:
    wanted_exception = walk_exception_context(
        exception, (requests.RequestException, urllib3.exceptions.HTTPError)
    )
    return exception if wanted_exception is None else wanted_exception


class HTTPEngineRequests:
    def __init__(self) -> None:
        self.session = requests.Session()

    def __call__(self, request: HTTPRequest) -> HTTPResponse:
        headers = request.headers

        # remove request/urllib3 User-Agent header
        if "user-agent" not in headers:
            headers = HTTPHeaderDict(headers)  # copy
            headers["user-agent"] = urllib3.util.SKIP_HEADER  # type: ignore[attr-defined]

        try:
            response = self.session.request(
                method=request.method,
                url=request.url,
                headers=headers,
                data=request.body,
                stream=request.stream_response,
                allow_redirects=False,
                timeout=(40, 600 if request.stream_response else 30),
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ProxyError,
        ) as ex:  # pragma: no cover
            raise HTTPConnectionError.from_exception(
                _extract_exception(ex), request=request
            ) from ex
        except requests.exceptions.Timeout as ex:  # pragma: no cover
            raise HTTPTimeoutError.from_exception(
                _extract_exception(ex), request=request
            ) from ex
        except Exception as ex:  # pragma: no cover  # noqa: BLE001
            raise HTTPError.from_exception(
                _extract_exception(ex), request=request
            ) from ex

        return HTTPResponseRequests(response)



from abc import ABC, abstractmethod
from base64 import b64encode
from copy import deepcopy
import sys
from typing import Any, cast

from sdkite.http.adapter import HTTPAdapter, HTTPAdapterSpec
from sdkite.http.model import HTTPRequest

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self


class Auth(ABC):
    name: str

    def __init__(
        self,
        http_adapter_spec: HTTPAdapterSpec,
        *,
        interceptor_order: int = 0,
    ) -> None:
        self._http_adapter_spec = http_adapter_spec
        self._interceptor_order = interceptor_order

    def __set_name__(self, _: Any, name: str) -> None:
        self.name = name
        self._http_adapter_spec.register_interceptor(
            "request_interceptor", self.name, self._interceptor_order
        )

    def __get__(self, instance: Any, __: Any) -> Self:
        if instance is None:
            return self
        attr_name = f"_sdkite_auth_{self.name}"
        try:
            instance_auth = getattr(instance, attr_name)
        except AttributeError:
            instance_auth = deepcopy(self)
            setattr(instance, attr_name, instance_auth)
        return cast(Self, instance_auth)

    @abstractmethod
    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        ...


class BasicAuth(Auth):
    def __init__(
        self,
        http_adapter_spec: HTTPAdapterSpec,
        username: str = "",
        password: str = "",
        *,
        interceptor_order: int = 0,
    ) -> None:
        super().__init__(http_adapter_spec, interceptor_order=interceptor_order)
        self.username = username
        self.password = password

    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        # pylint: disable=line-too-long

        # UTF-8 encoding seems to be the de-facto standard in browsers
        # see e.g. https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication#character_encoding_of_http_authentication

        # pylint: enable=line-too-long

        credentials = b64encode(f"{self.username}:{self.password}".encode()).decode()
        request.headers["Authorization"] = f"Basic {credentials}"
        return request


class NoAuth(Auth):
    def __call__(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        return request



from os import chdir, getcwd
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple, cast

import pytest
from requests_mock import Mocker

from sdkite import Client
from sdkite.http import HTTPRequest, HTTPResponse
import sdkite.http.adapter as adapter_module


@pytest.fixture(scope="module", autouse=True)
def _import_pytest(doctest_namespace: Dict[str, object]) -> None:
    doctest_namespace["Iterator"] = Iterator
    doctest_namespace["Path"] = Path

    doctest_namespace["Client"] = Client

    def list_spells(
        max_price: int,
        offset: int = 0,
        page: Optional[int] = None,
    ) -> List[str]:
        assert max_price == 50
        limit = 3
        if page is not None:
            offset = page * limit
        return [
            "Crushing Burden Touch",
            "Great Burden of Sin",
            "Heavy Burden",
            "Strong Feather",
            "Tinur's Hoptoad",
            "Ulms's Juicedaw's Feather",
            "Far Silence",
            "Soul Trap",
        ][offset:][:limit]

    def list_spells_with_ref(
        max_price: Optional[int] = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[str], Optional[str]]:
        assert max_price == 50
        cursors = [None, "57656c636", "f6d652074", "6f2042616", "c6d6f7261"]
        index = cursors.index(cursor)
        return list_spells(max_price, page=index), cursors[index + 1]

    class ExampleEngine:
        def __init__(self, *_: object, **__: object) -> None:
            pass

        def __call__(self, request: HTTPRequest) -> HTTPResponse:
            raise NotImplementedError

    doctest_namespace["list_spells"] = list_spells
    doctest_namespace["list_spells_with_ref"] = list_spells_with_ref
    doctest_namespace["ExampleEngine"] = ExampleEngine


@pytest.fixture(autouse=True)
def _patch_http(requests_mock: Mocker) -> None:
    #
    # quickstart
    #
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/world/npc/interact?name=ranis",
        text="Have you found the Telvanni spy?",
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/world/npc/interact?name=waldo",
        status_code=404,
    )
    requests_mock.register_uri(
        "POST",
        "https://api.example.com/world/npc/search",
        additional_matcher=lambda request: cast(
            bool,
            request.json() == {"city": "Balmora", "faction": "Mage", "min_level": 8},
        ),
        json=["marayn", "masalinie", "ranis"],
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/world/book/content?id=bk_words_of_the_wind",
        request_headers={"Authorization": "Basic TmVyZXZhcmluZTpJbmNhcm5hdGU="},
        text="Words of the Wind\n\nA volume of verse collected from...",
    )

    #
    # http_auth
    #
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/whoami",
        request_headers={"Authorization": "Basic QWxpY2U6VzBuZGVybEBuZA=="},
        text="Welcome Alice!",
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/noauth",
        additional_matcher=lambda request: "Authorization" not in request.headers,
        text="The /noauth endpoint has been called without auth",
    )

    #
    # http_response
    #
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/user/1",
        json={"name": "Alice", "age": 42},
    )
    requests_mock.register_uri(
        "GET",
        "https://api.example.com/user/2",
        json={"name": "Bob"},
    )


@pytest.fixture(autouse=True)
def _create_replay_store(tmp_path: Path) -> None:
    store = tmp_path / "replay"
    store.mkdir()
    (store / "ping.json").write_bytes(
        rb"""{
  "request": {
    "method": "GET",
    "url": "https://api.example.com/ping",
    "headers": {},
    "body": ""
  },
  "response": {
    "status_code": 200,
    "reason": "OK",
    "headers": {},
    "body": "pong"
  }
}
"""
    )


@pytest.fixture(autouse=True)
def _change_directory(tmp_path: Path) -> Iterator[None]:
    cwd = getcwd()
    try:
        chdir(tmp_path)
        yield
    finally:
        chdir(cwd)


@pytest.fixture(autouse=True)
def _patched_adapter(monkeypatch: pytest.MonkeyPatch) -> None:
    # change default value for faster tests
    monkeypatch.setattr(adapter_module, "_DEFAULT_WAIT_INITIAL", 0)
    monkeypatch.setattr(adapter_module, "_DEFAULT_WAIT_JITTER", 0)



from itertools import chain
import re
from typing import Iterable, List, Set, Tuple, Union

import pytest

from sdkite.http.utils import build_status_code_check

_CASES_DESCRIPTIONS: List[
    Tuple[str, Union[int, str, Iterable[Union[int, str]]], Iterable[int]]
] = [
    ("int", 200, [200]),
    ("str_exact", "200", [200]),
    ("str_2xx", "2xx", range(200, 300)),
    ("str_x00", "x00", range(0, 1000, 100)),
    ("list-int", [200, 300], [200, 300]),
    ("list-int-duplicates", [200, 200], [200]),
    ("list-str", ["200", "300"], [200, 300]),
    ("list-str-duplicates", ["200", "200"], [200]),
    (
        "list-str-separated-ranges",
        ["2xx", "4xx"],
        chain(range(200, 300), range(400, 500)),
    ),
    ("list-str-overlap", ["200", "2xx", "222"], range(200, 300)),
    ("list-mixed-exact", ["200", 300], [200, 300]),
    ("list-mixed-duplicates", ["200", 200], [200]),
]


@pytest.mark.parametrize(
    ["status_codes", "expected"],
    [
        pytest.param(status_codes, set(expected), id=name)
        for name, status_codes, expected in _CASES_DESCRIPTIONS
    ],
)
def test_valid_cases(
    status_codes: Union[int, str, Iterable[Union[int, str]]], expected: Set[int]
) -> None:
    checker = build_status_code_check(status_codes)
    matching = set(filter(checker, range(1000)))
    assert matching == expected


@pytest.mark.parametrize("status_codes", [-42, 1000, "-42", "1000", "4xy"])
def test_invalid(status_codes: Union[int, str, Iterable[Union[int, str]]]) -> None:
    with pytest.raises(
        ValueError,
        match=re.escape(r"Invalid status code (must match /^[0-9x]{3}$/): "),
    ):
        build_status_code_check(status_codes)



import sys
from typing import TYPE_CHECKING, Optional

from sdkite.exceptions import SDKiteError

if TYPE_CHECKING:  # pragma: no cover
    # avoid circular references
    from sdkite.http.model import HTTPRequest, HTTPResponse

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self


class HTTPError(SDKiteError):
    request: "HTTPRequest"
    response: Optional["HTTPResponse"]

    def __init__(
        self,
        *,
        msg: str = "N/A",
        request: "HTTPRequest",
        response: Optional["HTTPResponse"] = None,
    ) -> None:
        super().__init__(msg)
        self.request = request
        self.response = response

    @classmethod
    def from_exception(
        cls,
        exception: BaseException,
        *,
        request: "HTTPRequest",
        response: Optional["HTTPResponse"] = None,
    ) -> Self:
        return cls(
            msg=f"{exception.__class__.__name__}: {exception}",
            request=request,
            response=response,
        )


class HTTPConnectionError(HTTPError):
    pass


class HTTPTimeoutError(HTTPError):
    pass


class HTTPStatusCodeError(HTTPError):
    response: "HTTPResponse"

    def __init__(
        self,
        *,
        status_code: int,
        request: "HTTPRequest",
        response: Optional["HTTPResponse"],
    ) -> None:
        super().__init__(
            msg=f"Unexpected status code: {status_code}",
            request=request,
            response=response,
        )
        self.status_code = status_code


class HTTPContextError(HTTPError):
    response: "HTTPResponse"

    def __init__(
        self,
        *,
        msg: str = "N/A",
        request: "HTTPRequest",
        response: "HTTPResponse",
    ) -> None:
        super().__init__(msg=msg, request=request, response=response)



import re

import pytest

from sdkite.http._stringescape import stringescape_loads


@pytest.mark.parametrize(
    ["data", "error_msg"],
    [
        pytest.param(r"abc\x0", r"Invalid data to load ('\\x0')", id="truncated"),
        pytest.param(
            r"abc\u0123456789", r"Invalid data to load ('\\u01')", id="invalid-escape"
        ),
        pytest.param(
            "in the æther", "Invalid data to load (not ascii)", id="non-ascii"
        ),
    ],
)
def test_stringescape_loads_invalid(data: str, error_msg: str) -> None:
    with pytest.raises(ValueError, match=re.escape(error_msg)):
        stringescape_loads(data)



from pathlib import Path
import sys
from typing import Any, List, cast

import pytest
from requests_mock import Mocker

from sdkite import Client, Pagination, paginated
from sdkite.http import BasicAuth, HTTPAdapterSpec, HTTPRequest, NoAuth
from sdkite.http.engine_replay import HTTPEngineReplay, HTTPResponseReplay

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from typing import TypedDict


REPLAY_PATH = Path(__file__).parent / "recorded"


class ApiPublic(Client):
    _http = HTTPAdapterSpec(url="public")
    _auth = NoAuth(_http)

    def version(self) -> str:
        response = self._http.get("version")
        return response.data_str


class User(TypedDict):
    name: str
    age: int


class ApiUsers(Client):
    _http = HTTPAdapterSpec(url="users", headers={"X-Toto": "Abc"})

    def get(self, user_id: int) -> User:
        response = self._http.get(str(user_id))
        return cast(User, response.data_json)

    @paginated(page=1)
    def get_all(self, pagination: Pagination) -> List[User]:
        response = self._http.get(f"all/{pagination.page}")
        return cast(List[User], response.data_json)


class Api(Client):
    _http = HTTPAdapterSpec()
    _auth = BasicAuth(_http, username="tests")

    public: ApiPublic
    users: ApiUsers

    def __init__(self, url: str, password: str) -> None:
        super().__init__()
        self._http.url = url
        self._auth.password = password


@pytest.fixture
def _mock_http_requests(requests_mock: Mocker) -> None:
    common_response: Any = {
        "reason": "OK",
        "headers": {"server": "nginx"},
    }

    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/api/v1/public/version",
        additional_matcher=lambda request: "Authorization" not in request.headers,
        # response
        content=b"1.2.3",
        **common_response,
    )

    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/api/v1/users/1337",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        # response
        content=b'{"name":"John Doe","age":42}',
        **common_response,
    )

    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/api/v1/users/all/1",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        # response
        content=b'[{"name":"John Doe","age":42},{"name":"Alice Doe","age":41}]',
        **common_response,
    )
    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/api/v1/users/all/2",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        # response
        content=b'[{"name":"Bob Doe","age":10},{"name":"Carole Doe","age":12}]',
        **common_response,
    )
    requests_mock.register_uri(
        # request
        "GET",
        "https://www.example.com/api/v1/users/all/3",
        request_headers={"authorization": "Basic dGVzdHM6czNjcjN0"},
        # response
        content=b"[]",
        **common_response,
    )


@pytest.mark.usefixtures("_mock_http_requests")
def test_http() -> None:
    client = Api("https://www.example.com/api/v1", "s3cr3t")

    assert client.public.version() == "1.2.3"

    user = client.users.get(1337)
    assert user == User(name="John Doe", age=42)

    users = list(client.users.get_all())
    assert users == [
        User(name="John Doe", age=42),
        User(name="Alice Doe", age=41),
        User(name="Bob Doe", age=10),
        User(name="Carole Doe", age=12),
    ]


def replay_request_modifier(request: HTTPRequest) -> HTTPRequest:
    if "Authorization" in request.headers:
        request.headers["Authorization"] = "Basic redacted"
    return request


def test_http_replay_replay(
    # notice how we don't use `_mock_http_requests` here
    # still patching requests just in case
    requests_mock: Mocker,  # pylint: disable=unused-argument # noqa: ARG001
) -> None:
    Api._http.set_engine(  # pylint: disable=protected-access
        HTTPEngineReplay,
        [REPLAY_PATH],
        replay_request_modifier=replay_request_modifier,
    )

    client = Api(
        "https://www.example.com/api/v1",
        "whatever",  # password is ignored
    )

    assert client.public.version() == "1.2.3"

    user = client.users.get(1337)
    assert user == User(name="John Doe", age=42)

    users = list(client.users.get_all())
    assert users == [
        User(name="John Doe", age=42),
        User(name="Alice Doe", age=41),
        User(name="Bob Doe", age=10),
        User(name="Carole Doe", age=12),
    ]


@pytest.mark.usefixtures("_mock_http_requests")
def test_http_replay_record(tmp_path: Path) -> None:
    def recording_response_modifier(response: HTTPResponseReplay) -> HTTPResponseReplay:
        return response.replace(
            headers={
                **response.headers,
                "server": "apache",
            }
        )

    def recording_compute_basename(request: HTTPRequest, _: HTTPResponseReplay) -> str:
        prefix = "https://www.example.com/api/v1/"
        assert request.url.startswith(prefix)
        return request.url[len(prefix) :].replace("/", "_")

    Api._http.set_engine(  # pylint: disable=protected-access
        HTTPEngineReplay,
        [tmp_path],
        recording=True,
        replay_request_modifier=replay_request_modifier,
        recording_response_modifier=recording_response_modifier,
        recording_compute_basename=recording_compute_basename,
    )

    client = Api("https://www.example.com/api/v1", "s3cr3t")

    assert client.public.version() == "1.2.3"

    user = client.users.get(1337)
    assert user == User(name="John Doe", age=42)

    users = list(client.users.get_all())
    assert users == [
        User(name="John Doe", age=42),
        User(name="Alice Doe", age=41),
        User(name="Bob Doe", age=10),
        User(name="Carole Doe", age=12),
    ]

    expected_paths = {path.relative_to(REPLAY_PATH) for path in REPLAY_PATH.rglob("*")}
    assert {
        path.relative_to(tmp_path) for path in tmp_path.rglob("*")
    } == expected_paths

    for path in expected_paths:
        assert (tmp_path / path).read_bytes() == (REPLAY_PATH / path).read_bytes(), path



import sys

import pytest

from sdkite.http import HTTPContextError, HTTPHeaderDict, HTTPRequest, HTTPResponse

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterator
else:  # pragma: no cover
    from collections.abc import Iterator


class FakeResponse(HTTPResponse):
    def __init__(self) -> None:
        self.is_closed = False

    @property
    def raw(self) -> object:
        raise NotImplementedError

    @property
    def status_code(self) -> int:
        raise NotImplementedError

    @property
    def reason(self) -> str:
        raise NotImplementedError

    @property
    def headers(self) -> HTTPHeaderDict:
        raise NotImplementedError

    @property
    def data_stream(self) -> Iterator[bytes]:
        raise NotImplementedError

    @property
    def data_bytes(self) -> bytes:
        raise NotImplementedError

    @property
    def data_str(self) -> str:
        raise NotImplementedError

    @property
    def data_json(self) -> object:
        raise NotImplementedError

    def _close(self) -> None:
        self.is_closed = True


@pytest.mark.parametrize("raise_exception", [False, True])
def test_response_context_manager(raise_exception: bool) -> None:
    request = HTTPRequest(
        method="GET",
        url="https://example.com/",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    response = FakeResponse()
    response._set_context(request)  # pylint: disable=protected-access

    if raise_exception:
        with pytest.raises(HTTPContextError) as excinfo:  # noqa: SIM117, PT012
            with response:
                assert not response.is_closed
                raise ValueError("oops")

        exception = excinfo.value
        assert str(exception) == "ValueError: oops"
        assert exception.request == request
        assert exception.response == response
        assert isinstance(exception.__context__, ValueError)

    else:
        with response:
            assert not response.is_closed

    assert response.is_closed



import re
import sys
from typing import Any, Dict, List, Optional, Tuple, Union
from unittest.mock import Mock, call

import pytest

from sdkite.http import (
    HTTPAdapter,
    HTTPBodyEncoding,
    HTTPContextError,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPRequestAttemptInfo,
    HTTPResponse,
    HTTPStatusCodeError,
)

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterator
else:  # pragma: no cover
    from collections.abc import Iterator


class FakeResponse(HTTPResponse):
    def __init__(self, status_code: int = 200) -> None:
        self._status_code = status_code

    @property
    def raw(self) -> object:
        raise NotImplementedError

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def reason(self) -> str:
        raise NotImplementedError

    @property
    def headers(self) -> HTTPHeaderDict:
        raise NotImplementedError

    @property
    def data_stream(self) -> Iterator[bytes]:
        raise NotImplementedError

    @property
    def data_bytes(self) -> bytes:
        raise NotImplementedError

    @property
    def data_str(self) -> str:
        raise NotImplementedError

    @property
    def data_json(self) -> object:
        raise NotImplementedError


def create_adapter(
    url: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    parent_adapter: Optional[HTTPAdapter] = None,
) -> Tuple[HTTPAdapter, Mock, Mock]:
    # pylint: disable=protected-access
    send_request = Mock()
    send_request.return_value.status_code = 200
    adapter = HTTPAdapter(send_request)
    adapter._attr_name = "http"
    client = Mock()
    client.http = adapter
    adapter._clients = (client,)
    if parent_adapter:
        adapter._clients = parent_adapter._clients + adapter._clients
    adapter.url = url
    adapter.headers = HTTPHeaderDict(headers)
    adapter.retry_nb_attempts = None
    adapter.retry_callback = None
    adapter.retry_wait_initial = 0  # change default value for faster tests
    adapter.retry_wait_max = None
    adapter.retry_wait_jitter = 0  # change default value for faster tests
    adapter.request_interceptor = {}
    adapter.response_interceptor = {}
    return (adapter, send_request, client)


@pytest.mark.parametrize("method", ["post", "pOsT", "POST"])
def test_request_method(method: str) -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request(method, "https://www.example.com")
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            )
        )
    ]


@pytest.mark.parametrize(
    ["base_url", "url"],
    [
        (None, "https://www.example.com/foo/bar"),
        ("https://www.example.com", "foo/bar"),
        ("https://www.example.com/", "foo/bar"),
        ("https://www.example.com/foo", "bar"),
        ("https://www.example.com/foo/", "bar"),
        ("https://www.example.com/foo/bar", ""),
        ("https://www.example.com/foo/bar", None),
        ("https://www.example.com", "/foo/bar"),
        ("https://www.example.com/", "/foo/bar"),
        ("https://www.example.com/abc/def", "/foo/bar"),
        ("https://www.example.com/foo/abc/def", "../../bar"),
    ],
)
def test_request_url(base_url: Optional[str], url: Optional[str]) -> None:
    adapter, send_request, _ = create_adapter(base_url)
    response = adapter.request("GET", url)
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com/foo/bar",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            )
        )
    ]


def test_request_retry_status_code() -> None:
    adapter, send_request, _ = create_adapter()

    request_attempt = 0

    def do_send_request(_: HTTPRequest) -> FakeResponse:
        nonlocal request_attempt
        request_attempt += 1
        return FakeResponse(200 if request_attempt == 3 else 503)

    send_request.side_effect = do_send_request

    # fail with only 2 attempts
    with pytest.raises(HTTPStatusCodeError):
        adapter.request("GET", "https://www.example.com", retry_nb_attempts=2)

    # works with more attempts
    request_attempt = 0
    adapter.request("GET", "https://www.example.com")


@pytest.mark.parametrize(
    ["success_at_attempt_nb", "retry_nb_attempts_arg", "expected_error_at_attempt"],
    [
        # default retry_nb_attempts
        pytest.param(1, None, None, id="default-1"),
        pytest.param(2, None, None, id="default-2"),
        pytest.param(3, None, None, id="default-3"),
        pytest.param(4, None, 3, id="default-4"),
        pytest.param(5, None, 3, id="default-5"),
        # custom retry_nb_attempts argument
        pytest.param(1, 2, None, id="custom-1"),
        pytest.param(2, 2, None, id="custom-2"),
        pytest.param(3, 2, 2, id="custom-3"),
        pytest.param(4, 2, 2, id="custom-4"),
        pytest.param(5, 2, 2, id="custom-5"),
    ],
)
def test_request_retry(
    success_at_attempt_nb: int,
    retry_nb_attempts_arg: Optional[int],
    expected_error_at_attempt: Optional[bool],
) -> None:
    adapter, send_request, client = create_adapter()

    adapter.request_interceptor["inter"] = 0

    # side-effect on request in interceptor to test initial_request in retry callback
    def intercept(request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
        request.headers["X-Intercepted"] = "yes"
        return request

    client.inter.side_effect = intercept

    expected_response = FakeResponse()

    class CustomError(Exception):
        def __eq__(self, other: Any) -> bool:
            return type(self) is type(other) and self.args == other.args

    request_attempt = 0

    def do_send_request(_: HTTPRequest) -> FakeResponse:
        nonlocal request_attempt
        request_attempt += 1
        if request_attempt == success_at_attempt_nb:
            return expected_response
        raise CustomError(f"Send request error {request_attempt}")

    send_request.side_effect = do_send_request

    logged_retries: List[Tuple[int, BaseException]] = []

    def retry_callback(attempt_info: HTTPRequestAttemptInfo) -> None:
        logged_retries.append((attempt_info.attempt_number, attempt_info.exception))
        assert attempt_info.initial_request == HTTPRequest(
            "GET", "https://www.example.com", HTTPHeaderDict(), b"", False
        )
        assert 0.0 < attempt_info.seconds_since_start < 0.1

    kwargs: Any = {
        "retry_callback": retry_callback,
    }
    if retry_nb_attempts_arg is not None:
        kwargs["retry_nb_attempts"] = retry_nb_attempts_arg

    if expected_error_at_attempt is None:
        response = adapter.request("GET", "https://www.example.com", **kwargs)
        assert response == expected_response

    else:
        with pytest.raises(
            CustomError,
            match=re.escape(f"Send request error {expected_error_at_attempt}"),
        ):
            adapter.request("GET", "https://www.example.com", **kwargs)

    assert logged_retries == [
        (i, CustomError(f"Send request error {i}"))
        for i in range(1, min(success_at_attempt_nb, retry_nb_attempts_arg or 3))
    ], "retry_callback"


def test_no_request_url() -> None:
    adapter, _, _ = create_adapter()
    with pytest.raises(ValueError, match=re.escape("No URL provided")):
        adapter.request("GET")


def test_request_body_raw() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request("POST", "https://www.example.com", body=b"foobar")
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"foobar",
                stream_response=False,
            )
        )
    ]


def test_request_body_json() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request(
        "POST",
        "https://www.example.com",
        body="foobar",
        body_encoding=HTTPBodyEncoding.JSON,
    )
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="POST",
                url="https://www.example.com",
                headers=HTTPHeaderDict({"content-type": "application/json"}),
                body=b'"foobar"',
                stream_response=False,
            )
        )
    ]


def test_request_headers() -> None:
    adapter, send_request, _ = create_adapter(
        headers={
            "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
            "referer": "https://www.example.com",
        }
    )

    response = adapter.request(
        "GET",
        "https://www.example.com",
        headers={"referer": "https://www.example.com/foo/bar"},
    )
    assert response == send_request.return_value

    response = adapter.request(
        "GET",
        "https://www.example.com",
        headers={"x-foo": "bar"},
    )
    assert response == send_request.return_value

    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(
                    {
                        "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
                        "referer": "https://www.example.com/foo/bar",
                    }
                ),
                body=b"",
                stream_response=False,
            )
        ),
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(
                    {
                        "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
                        "referer": "https://www.example.com",
                        "x-foo": "bar",
                    }
                ),
                body=b"",
                stream_response=False,
            )
        ),
    ]


def test_request_stream_response() -> None:
    adapter, send_request, _ = create_adapter()
    response = adapter.request("GET", "https://www.example.com", stream_response=True)
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=True,
            )
        )
    ]


def test_request_with_interceptors() -> None:
    adapter, send_request, client = create_adapter()
    adapter.request_interceptor["inter0"] = 0
    adapter.request_interceptor["inter1"] = 1
    adapter.response_interceptor["inter2"] = 1
    adapter.response_interceptor["inter3"] = 0

    client.inter2.return_value.status_code = 200

    response = adapter.request("GET", "https://www.example.com")

    assert client.inter0.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            ),
            adapter,
        )
    ]
    assert client.inter1.call_args_list == [
        call(
            client.inter0.return_value,
            adapter,
        )
    ]
    assert send_request.call_args_list == [
        call(
            client.inter1.return_value,
        )
    ]
    assert client.inter3.call_args_list == [
        call(
            send_request.return_value,
            adapter,
        )
    ]
    assert client.inter2.call_args_list == [
        call(
            client.inter3.return_value,
            adapter,
        )
    ]
    assert response == client.inter2.return_value


def test_request_interceptor_crash() -> None:
    adapter, _, client = create_adapter()
    adapter.response_interceptor["inter"] = 0

    class CustomError(Exception):
        pass

    client.inter.side_effect = CustomError("Interceptor crash msg")

    with pytest.raises(CustomError, match=re.escape("Interceptor crash msg")):
        adapter.request("GET", "https://www.example.com")


def test_request_with_interceptors_hierarchy() -> None:
    adapter0, _, client0 = create_adapter()
    adapter1, _, client1 = create_adapter(parent_adapter=adapter0)
    adapter2, send_request2, client2 = create_adapter(parent_adapter=adapter1)

    adapter0.request_interceptor["inter0"] = 0
    adapter0.request_interceptor["inter1"] = -1
    adapter0.request_interceptor["inter2"] = -2
    adapter1.request_interceptor["inter1"] = 1
    adapter2.request_interceptor["inter2"] = 2

    response = adapter2.request("GET", "https://www.example.com")

    assert client0.inter0.call_args_list == [
        call(
            HTTPRequest(
                method="GET",
                url="https://www.example.com",
                headers=HTTPHeaderDict(),
                body=b"",
                stream_response=False,
            ),
            adapter2,
        )
    ]
    assert client1.inter1.call_args_list == [
        call(
            client0.inter0.return_value,
            adapter2,
        )
    ]
    assert client2.inter2.call_args_list == [
        call(
            client1.inter1.return_value,
            adapter2,
        )
    ]
    assert send_request2.call_args_list == [
        call(
            client2.inter2.return_value,
        )
    ]
    assert response == send_request2.return_value


@pytest.mark.parametrize(
    "method", ["get", "options", "head", "post", "put", "patch", "delete"]
)
def test_adapter_request_sugar(method: str) -> None:
    adapter, send_request, _ = create_adapter()
    request_with_method = getattr(adapter, method)

    response = request_with_method(
        "https://www.example.com/foo/bar",
        body="foobar",
        body_encoding=HTTPBodyEncoding.JSON,
        headers={
            "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
            "referer": "https://www.example.com/",
        },
        stream_response=True,
    )
    assert response == send_request.return_value
    assert send_request.call_args_list == [
        call(
            HTTPRequest(
                method=method.upper(),
                url="https://www.example.com/foo/bar",
                headers=HTTPHeaderDict(
                    {
                        "authorization": "Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==",
                        "referer": "https://www.example.com/",
                        "content-type": "application/json",
                    }
                ),
                body=b'"foobar"',
                stream_response=True,
            )
        )
    ]


@pytest.mark.parametrize("status_code", [204, 301, 404, 500])
def test_invalid_status_code(status_code: int) -> None:
    adapter, send_request, _ = create_adapter()
    send_request.return_value.status_code = status_code

    with pytest.raises(
        HTTPStatusCodeError, match=f"^Unexpected status code: {status_code}$"
    ) as excinfo:
        adapter.request("GET", "https://www.example.com")
    assert excinfo.value.status_code == status_code

    with pytest.raises(
        HTTPStatusCodeError, match=f"^Unexpected status code: {status_code}$"
    ) as excinfo:
        adapter.get("https://www.example.com")
    assert excinfo.value.status_code == status_code

    expected_status_codes_list: List[Union[int, str, Tuple[int, int]]] = [
        status_code,
        str(status_code),
        "xxx",
        (999, status_code),
    ]

    for expected_status_codes in expected_status_codes_list:
        adapter.request(
            "GET",
            "https://www.example.com",
            expected_status_codes=expected_status_codes,
        )
        adapter.get(
            "https://www.example.com",
            expected_status_codes=expected_status_codes,
        )


def test_adapter_context_manager() -> None:
    adapter, send_request, _ = create_adapter()

    def do_send_request(_: HTTPRequest) -> HTTPResponse:
        return FakeResponse()

    send_request.side_effect = do_send_request

    with adapter.get("https://www.example.com") as response:
        assert isinstance(response, FakeResponse)

    with pytest.raises(HTTPContextError) as excinfo:  # noqa: PT012,SIM117
        with adapter.get("https://www.example.com") as response:
            raise ValueError("oops")

    exception = excinfo.value
    assert str(exception) == "ValueError: oops"
    assert exception.request == HTTPRequest(
        method="GET",
        url="https://www.example.com",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    assert isinstance(exception.response, FakeResponse)
    assert isinstance(exception.__context__, ValueError)



import re
import sys
from typing import List, Optional

import pytest

from sdkite.utils import identity, last_not_none, walk_exception_context, zip_reverse

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import assert_type
else:  # pragma: no cover
    from typing import assert_type


def test_identity() -> None:
    assert identity(42) == 42
    obj = object()
    assert identity(obj) == obj


def test_zip_reverse() -> None:
    assert list(zip_reverse((1, 2), "ab")) == [(2, "b"), (1, "a")]


def test_zip_reverse_invalid_sizes() -> None:
    with pytest.raises(
        ValueError,
        match=re.escape("zip_reverse() arguments have different lengths"),
    ):
        zip_reverse((1, 2), "abc")


@pytest.mark.parametrize(
    ["items", "expected"],
    [
        pytest.param([], None, id="empty"),
        pytest.param([1], 1, id="mandatory-1"),
        pytest.param([1, 2, 3], 3, id="mandatory-3"),
        pytest.param([None, None, 1, 2, 3], 3, id="optional-before"),
        pytest.param([1, 2, 3, None, None], 3, id="optional-after"),
        pytest.param([None, 1, None, 2, None, 3, None], 3, id="optional-interleaved"),
    ],
)
def test_last_not_none(items: List[Optional[int]], expected: Optional[int]) -> None:
    assert last_not_none(items) == expected
    if expected is None:
        assert last_not_none(items, -42) == -42
    else:
        assert last_not_none(items, -42) == expected


def test_last_not_none_typing() -> None:
    mandatory_list: List[int] = []
    optional_list: List[Optional[int]] = []

    assert_type(last_not_none(mandatory_list), Optional[int])
    assert_type(last_not_none(mandatory_list, 42), int)
    assert_type(last_not_none(optional_list), Optional[int])
    assert_type(last_not_none(optional_list, 42), int)


@pytest.mark.parametrize("exception_cause", [None, True, False])
def test_walk_exception_context(exception_cause: Optional[bool]) -> None:
    class AException(BaseException):
        pass

    class BException(BaseException):
        pass

    class CException(BaseException):
        pass

    exception = None
    try:
        try:
            try:
                raise AException
            except AException as exa:
                raise BException from exa
        except BException as exb:
            if exception_cause is None:
                raise CException from None
            if exception_cause:
                raise CException from exb
            raise CException  # pylint: disable=raise-missing-from  # noqa: B904
    except CException as cex:
        exception = cex

    assert isinstance(exception, CException)

    assert isinstance(
        walk_exception_context(exception, ValueError),
        CException,
    )
    assert isinstance(
        walk_exception_context(exception, CException),
        BException,
    )
    assert isinstance(
        walk_exception_context(exception, (BException, CException)),
        AException,
    )
    assert (
        walk_exception_context(exception, (AException, BException, CException)) is None
    )



from pathlib import Path
import re
import time
from typing import Set

import pytest
from requests_mock import Mocker, NoMockAddress

from sdkite.http import HTTPError, HTTPHeaderDict, HTTPRequest
from sdkite.http.engine_replay import HTTPEngineReplay, HTTPResponseReplay

REPLAY_PATH = Path(__file__).parent / "engine_replay"


@pytest.fixture(autouse=True)
def _patched_time(monkeypatch: pytest.MonkeyPatch) -> None:
    def mocked_time() -> int:
        return 1234567890

    monkeypatch.setattr(time, "time", mocked_time)


def list_files(src: Path) -> Set[Path]:
    return {path.relative_to(src) for path in src.rglob("*") if path.is_file()}


def test_recording(tmp_path: Path, requests_mock: Mocker) -> None:
    original_get_root_data = (REPLAY_PATH / "base" / "get_root.json").read_bytes()
    original_get_foo_data = (REPLAY_PATH / "base" / "get_foo.json").read_bytes()
    (tmp_path / "aaa").mkdir(parents=True)
    (tmp_path / "bbb").mkdir(parents=True)
    (tmp_path / "aaa" / "get_root.json").write_bytes(original_get_root_data)
    (tmp_path / "aaa" / "get_foo.json").write_bytes(original_get_foo_data)

    engine = HTTPEngineReplay([tmp_path / "aaa", tmp_path / "bbb"], recording=True)

    requests_mock.register_uri(
        # request
        "GET",
        "https://example.com/foo",
        # response
        content=b'{"msg": 13}',
        reason="OK",
        headers={"server": "nginx"},
    )
    requests_mock.register_uri(
        # request
        "GET",
        "https://example.com/bar",
        # response
        content=b'{"msg": 37}',
        reason="OK",
        headers={"server": "nginx"},
    )

    # missing
    request = HTTPRequest(
        method="GET",
        url="https://example.com/bar",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    response = engine(request)
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.headers == HTTPHeaderDict({"server": "nginx"})
    assert response.data_json == {"msg": 37}
    assert list_files(tmp_path) == {
        Path() / "aaa" / "get_root.json",
        Path() / "aaa" / "get_foo.json",
        Path() / "bbb" / "1234567890_get_https___example_com_bar.json",
    }
    assert (
        (tmp_path / "bbb" / "1234567890_get_https___example_com_bar.json").read_bytes()
        == rb"""{
  "request": {
    "method": "GET",
    "url": "https://example.com/bar",
    "headers": {},
    "body": ""
  },
  "response": {
    "status_code": 200,
    "reason": "OK",
    "headers": {
      "server": "nginx"
    },
    "body": [
      "{\"msg\": 37}"
    ]
  }
}
"""
    )

    # existing
    request = HTTPRequest(
        method="GET",
        url="https://example.com/foo",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    response = engine(request)
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.headers == HTTPHeaderDict({"server": "nginx"})
    assert response.data_json == {"msg": 13}
    assert list_files(tmp_path) == {
        Path() / "aaa" / "get_root.json",
        Path() / "aaa" / "get_foo.json",
        Path() / "bbb" / "1234567890_get_https___example_com_bar.json",
        Path() / "bbb" / "1234567890_get_https___example_com_foo.json",
    }
    assert (tmp_path / "aaa" / "get_foo.json").read_bytes() == original_get_foo_data
    assert (
        (tmp_path / "bbb" / "1234567890_get_https___example_com_foo.json").read_bytes()
        == rb"""{
  "request": {
    "method": "GET",
    "url": "https://example.com/foo",
    "headers": {},
    "body": ""
  },
  "response": {
    "status_code": 200,
    "reason": "OK",
    "headers": {
      "server": "nginx"
    },
    "body": [
      "{\"msg\": 13}"
    ]
  }
}
"""
    )

    # crash
    request = HTTPRequest(
        method="GET",
        url="https://example.com/",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    with pytest.raises(
        HTTPError,
        match=re.escape("NoMockAddress: No mock address: GET https://example.com/"),
    ) as excinfo:
        response = engine(request)
    assert (  # pylint: disable=unidiomatic-typecheck
        type(excinfo.value.__context__) is NoMockAddress
    )


@pytest.mark.parametrize("recording", [True, False])
def test_recording_modifiers(
    tmp_path: Path, requests_mock: Mocker, recording: bool
) -> None:
    expected_recording = rb"""{
  "request": {
    "method": "GET",
    "url": "https://example.com/foo",
    "headers": {
      "Authorization": "Bearer fake_creds"
    },
    "body": ""
  },
  "response": {
    "status_code": 200,
    "reason": "OK",
    "headers": {
      "server": "nginx"
    },
    "body": [
      "{\"msg\": \"saved\"}"
    ]
  }
}
"""

    if recording:
        requests_mock.register_uri(
            # request
            "GET",
            "https://example.com/foo",
            request_headers={"authorization": "Bearer valid_creds"},
            # response
            content=b'{"msg": "real"}',
            reason="OK",
            headers={"server": "nginx"},
        )
    else:
        (tmp_path / "recorded.json").write_bytes(expected_recording)

    def replay_request_modifier(request: HTTPRequest) -> HTTPRequest:
        if "Authorization" in request.headers:
            request.headers["Authorization"] = request.headers["Authorization"].replace(
                "some_creds", "fake_creds"
            )
        return request

    response_id = -1

    def replay_response_modifier(response: HTTPResponseReplay) -> HTTPResponseReplay:
        nonlocal response_id
        response_id += 1
        return response.replace(
            body=response.data_bytes.replace(b"saved", b"modified%d" % response_id)
        )

    def recording_request_modifier(request: HTTPRequest) -> HTTPRequest:
        if "Authorization" in request.headers:
            request.headers["Authorization"] = request.headers["Authorization"].replace(
                "some_creds", "valid_creds"
            )
        return request

    def recording_response_modifier(response: HTTPResponseReplay) -> HTTPResponseReplay:
        return response.replace(body=response.data_bytes.replace(b"real", b"saved"))

    engine = HTTPEngineReplay(
        [tmp_path],
        recording=recording,
        replay_request_modifier=replay_request_modifier,
        replay_response_modifier=replay_response_modifier,
        recording_request_modifier=recording_request_modifier,
        recording_response_modifier=recording_response_modifier,
    )

    request = HTTPRequest(
        method="GET",
        url="https://example.com/foo",
        headers=HTTPHeaderDict({"Authorization": "Bearer some_creds"}),
        body=b"",
        stream_response=False,
    )
    response = engine(request)
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.headers == HTTPHeaderDict({"server": "nginx"})
    assert response.data_json == {"msg": "modified0"}

    # no mutations in request
    assert request == HTTPRequest(
        method="GET",
        url="https://example.com/foo",
        headers=HTTPHeaderDict({"Authorization": "Bearer some_creds"}),
        body=b"",
        stream_response=False,
    )

    if recording:
        assert (
            tmp_path / "1234567890_get_https___example_com_foo.json"
        ).read_bytes() == expected_recording

    # response unmodified
    response = engine(request)
    assert response.data_json == {"msg": "modified1"}



from contextlib import nullcontext
import re
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, call

import pytest

from sdkite.http import (
    HTTPAdapter,
    HTTPAdapterSendRequest,
    HTTPAdapterSpec,
    HTTPBodyEncoding,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPResponse,
)
from sdkite.http import adapter as adapter_module
from sdkite.http.adapter import _BeforeSleep

if TYPE_CHECKING:
    from sdkite import Client
else:
    # we want independent unit tests
    Client = object


class FakeResponse(HTTPResponse):
    def __init__(self, context: str, raw: object) -> None:
        self.context = context
        self._raw = raw

    @property
    def raw(self) -> object:
        return self._raw

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, FakeResponse)
            and self.context == other.context
            and self.raw == other.raw
        )

    def __repr__(self) -> str:
        return f"FakeResponse<{self.context}, {self.raw}>"

    status_code = 200
    reason = "Ok"
    headers = HTTPHeaderDict()
    data_stream = iter(())
    data_bytes = b""
    data_str = ""
    data_json = None


@pytest.fixture(autouse=True)
def _patched_adapter(monkeypatch: pytest.MonkeyPatch) -> None:
    engine = Mock()
    engine.return_value = lambda request: FakeResponse("send_request", request)
    monkeypatch.setattr(adapter_module, "HTTPEngineRequests", engine)


@pytest.fixture
def patched_tenacity(monkeypatch: pytest.MonkeyPatch) -> Mock:
    tenacity = Mock()

    tenacity.Retrying.return_value = [nullcontext()]
    tenacity.stop_after_attempt.side_effect = call
    tenacity.wait_exponential_jitter.side_effect = call

    for attr in ["Retrying", "stop_after_attempt", "wait_exponential_jitter"]:
        monkeypatch.setattr(adapter_module, attr, getattr(tenacity, attr))

    return tenacity


def test_base(
    patched_tenacity: Mock,  # pylint: disable=redefined-outer-name
) -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

    expected_request = HTTPRequest(
        method="GET",
        url="https://www.example.com/xxx/uvw",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse("send_request", expected_request)

    assert patched_tenacity.Retrying.call_args_list == [
        call(
            stop=patched_tenacity.stop_after_attempt(3),
            wait=patched_tenacity.wait_exponential_jitter(
                initial=1.0, max=60.0, jitter=1.0
            ),
            before_sleep=_BeforeSleep(None, expected_request),
            reraise=True,
        )
    ]


def test_retry_at_spec_level(
    patched_tenacity: Mock,  # pylint: disable=redefined-outer-name
) -> None:
    # these get deepcopy-ed so Mock() would not work
    retry_callback0: Any = "retry_callback0"
    retry_callback1: Any = "retry_callback1"

    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(
            url="https://www.example.com/xxx",
            retry_nb_attempts=1,
            retry_wait_initial=2.0,
            retry_wait_max=3.0,
            retry_wait_jitter=4.0,
            retry_callback=retry_callback0,
        )

    expected_request = HTTPRequest(
        method="GET",
        url="https://www.example.com/xxx/uvw",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )

    client = Klass()

    client.xxx.request("GET", "uvw")
    assert patched_tenacity.Retrying.call_args_list == [
        call(
            stop=patched_tenacity.stop_after_attempt(1),
            wait=patched_tenacity.wait_exponential_jitter(
                initial=2.0, max=3.0, jitter=4.0
            ),
            before_sleep=_BeforeSleep(retry_callback0, expected_request),
            reraise=True,
        )
    ]
    patched_tenacity.Retrying.reset_mock()

    client.xxx.request(
        "GET",
        "uvw",
        retry_nb_attempts=6,
        retry_wait_initial=7.0,
        retry_wait_max=8.0,
        retry_wait_jitter=9.0,
        retry_callback=retry_callback1,
    )
    assert patched_tenacity.Retrying.call_args_list == [
        call(
            stop=patched_tenacity.stop_after_attempt(6),
            wait=patched_tenacity.wait_exponential_jitter(
                initial=7.0, max=8.0, jitter=9.0
            ),
            before_sleep=_BeforeSleep(retry_callback1, expected_request),
            reraise=True,
        )
    ]


def test_overidden_content_type() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

    client = Klass()
    with pytest.warns(
        UserWarning,
        match=re.escape(
            "The 'content-type' header is being overridden due to request body encoding"
            " HTTPBodyEncoding.JSON (from 'custom' to 'application/json')"
        ),
    ):
        response = client.xxx.request(
            "GET",
            headers={"Content-Type": "custom"},
            body_encoding=HTTPBodyEncoding.JSON,
        )
    assert response == FakeResponse(
        "send_request",
        HTTPRequest(
            method="GET",
            url="https://www.example.com/xxx",
            headers=HTTPHeaderDict({"content-type": "application/json"}),
            body=b"null",
            stream_response=False,
        ),
    )


def test_interceptor_method() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        def xxx_req(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        def xxx_resp(self, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "xxx_resp",
        FakeResponse(
            "send_request",
            HTTPRequest(
                method="GET",
                url="https://www.example.com/xxx/uvw",
                headers=HTTPHeaderDict({"intercept": "xxx"}),
                body=b"",
                stream_response=False,
            ),
        ),
    )


def test_interceptor_staticmethod() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")
        yyy = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        @staticmethod
        def xxx_req(request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        @staticmethod
        def xxx_resp(response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

        @staticmethod
        @yyy.intercept_request(0)
        def yyy_intercept(request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "yyy")
            return request

        @staticmethod
        @yyy.intercept_response(1)
        def yyy_resp(response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("yyy_resp", response)

    client = Klass()
    for attr in ("xxx", "yyy"):
        response = getattr(client, attr).request("GET", "uvw")
        assert response == FakeResponse(
            f"{attr}_resp",
            FakeResponse(
                "send_request",
                HTTPRequest(
                    method="GET",
                    url="https://www.example.com/xxx/uvw",
                    headers=HTTPHeaderDict({"intercept": attr}),
                    body=b"",
                    stream_response=False,
                ),
            ),
        )


def test_interceptor_classmethod() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")
        yyy = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        @classmethod
        def xxx_req(cls, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        @classmethod
        def xxx_resp(cls, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

        @classmethod
        @yyy.intercept_request(0)
        def yyy_intercept(cls, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "yyy")
            return request

        @classmethod
        @yyy.intercept_response(1)
        def yyy_resp(cls, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("yyy_resp", response)

    client = Klass()
    for attr in ("xxx", "yyy"):
        response = getattr(client, attr).request("GET", "uvw")
        assert response == FakeResponse(
            f"{attr}_resp",
            FakeResponse(
                "send_request",
                HTTPRequest(
                    method="GET",
                    url="https://www.example.com/xxx/uvw",
                    headers=HTTPHeaderDict({"intercept": attr}),
                    body=b"",
                    stream_response=False,
                ),
            ),
        )


def test_register_interceptor_existing() -> None:
    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

        @xxx.intercept_request(0)
        def xxx_req(self, request: HTTPRequest, _: HTTPAdapter) -> HTTPRequest:
            request.headers.add("intercept", "xxx")
            return request

        @xxx.intercept_response(1)
        def xxx_resp(self, response: HTTPResponse, _: HTTPAdapter) -> HTTPResponse:
            return FakeResponse("xxx_resp", response)

    with pytest.warns(
        UserWarning,
        match=re.escape(
            "Interceptor 'xxx_req' of 'xxx' has already been registered"
            " with order 0, ignoring new registration with order 42"
        ),
    ):
        Klass.xxx.register_interceptor("request_interceptor", "xxx_req", 42)

    with pytest.warns(
        UserWarning,
        match=re.escape(
            "Interceptor 'xxx_resp' of 'xxx' has already been registered"
            " with order 1, ignoring new registration with order 42"
        ),
    ):
        Klass.xxx.register_interceptor("response_interceptor", "xxx_resp", 42)

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "xxx_resp",
        FakeResponse(
            "send_request",
            HTTPRequest(
                method="GET",
                url="https://www.example.com/xxx/uvw",
                headers=HTTPHeaderDict({"intercept": "xxx"}),  # only once
                body=b"",
                stream_response=False,
            ),
        ),
    )


def test_custom_engine() -> None:
    def custom_engine(param0: int, *, param1: int) -> HTTPAdapterSendRequest:
        def send_request(request: HTTPRequest) -> HTTPResponse:
            return FakeResponse(f"engine-{param0}-{param1}", request)

        return send_request

    class Klass(Client):
        _parent = None

        xxx = HTTPAdapterSpec(url="https://www.example.com/xxx")

    with pytest.raises(TypeError):
        Klass.xxx.set_engine(custom_engine)  # type: ignore[call-arg]
    with pytest.raises(TypeError):
        Klass.xxx.set_engine(custom_engine, 42)  # type: ignore[call-arg]
    with pytest.raises(TypeError):
        Klass.xxx.set_engine(custom_engine, 42, 1337)  # type: ignore[misc]
    with pytest.raises(TypeError):
        Klass.xxx.set_engine(custom_engine, param1=1337)  # type: ignore[call-arg]

    Klass.xxx.set_engine(custom_engine, 42, param1=1337)

    client = Klass()
    response = client.xxx.request("GET", "uvw")
    assert response == FakeResponse(
        "engine-42-1337",
        HTTPRequest(
            method="GET",
            url="https://www.example.com/xxx/uvw",
            headers=HTTPHeaderDict(),
            body=b"",
            stream_response=False,
        ),
    )



import sys
from typing import Optional, Tuple, Type, TypeVar, Union, overload

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Reversible, Sequence
else:  # pragma: no cover
    from collections.abc import Iterable, Reversible, Sequence


T = TypeVar("T")
U = TypeVar("U")


def identity(value: T) -> T:
    return value


def zip_reverse(items_a: Sequence[T], items_b: Sequence[U]) -> Iterable[Tuple[T, U]]:
    if len(items_a) != len(items_b):
        # in Python >= 3.10 we could use zip(..., strict=True)
        # but since we have sequences anyways this avoids dealing with backward compatibility
        raise ValueError("zip_reverse() arguments have different lengths")
    return zip(reversed(items_a), reversed(items_b))


@overload
def last_not_none(items: Reversible[T]) -> Optional[T]:
    ...


@overload
def last_not_none(items: Reversible[Optional[T]], default: T) -> T:
    ...


def last_not_none(
    items: Reversible[Optional[T]], default: Optional[T] = None
) -> Optional[T]:
    for item in reversed(items):
        if item is not None:
            return item
    return default


def walk_exception_context(
    exception: Optional[BaseException],
    exclude_type: Union[Type[BaseException], Tuple[Type[BaseException], ...]],
) -> Optional[BaseException]:
    while exception is not None:
        if not isinstance(exception, exclude_type):
            return exception
        exception = exception.__context__
    return None



from dataclasses import replace
from pathlib import Path
import re
from typing import Dict, cast

import pytest

from sdkite.http import HTTPHeaderDict, HTTPRequest
from sdkite.http.engine_replay import (
    HTTPEngineReplay,
    HTTPResponseReplay,
    _RecordedResponse,
)

REPLAY_PATH = Path(__file__).parent / "engine_replay"


def test_response() -> None:
    recorded_response: _RecordedResponse = {
        "status_code": 200,
        "reason": "OK",
        "headers": {"content-type": "application/json"},
        "body": [b'{"msg":', b" 4", b"2}"],
    }
    expected_headers = HTTPHeaderDict({"content-type": "application/json"})

    response_replay = HTTPResponseReplay(recorded_response)
    with pytest.raises(
        ValueError,
        match=re.escape("The 'raw' attribute is not available with the replay engine"),
    ):
        response_replay.raw  # pylint: disable=pointless-statement  # noqa: B018
    assert response_replay.status_code == 200
    assert response_replay.reason == "OK"
    assert response_replay.headers == expected_headers

    response_replay = HTTPResponseReplay(recorded_response)
    data_stream = response_replay.data_stream
    assert next(data_stream) == b'{"msg":'
    assert next(data_stream) == b" 4"
    assert next(data_stream) == b"2}"
    assert next(data_stream, None) is None

    response_replay = HTTPResponseReplay(recorded_response)
    assert next(response_replay.data_stream) == b'{"msg":'
    assert next(response_replay.data_stream) == b" 4"
    assert next(response_replay.data_stream) == b"2}"
    assert next(response_replay.data_stream, None) is None

    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.data_bytes == b'{"msg": 42}'
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The data_xxx attributes can be only accessed once with the replay engine"
        ),
    ):
        response_replay.data_bytes  # pylint: disable=pointless-statement  # noqa: B018

    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.data_str == '{"msg": 42}'
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The data_xxx attributes can be only accessed once with the replay engine"
        ),
    ):
        response_replay.data_str  # pylint: disable=pointless-statement  # noqa: B018

    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.data_json == {"msg": 42}
    assert response_replay.data_json == {"msg": 42}

    # modify headers
    response_replay = HTTPResponseReplay(recorded_response)
    response_replay.headers["content-type"] = "new"
    assert response_replay.headers["content-type"] == "new"
    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.headers == expected_headers

    # modify json
    response_replay = HTTPResponseReplay(recorded_response)
    data = cast(Dict[str, int], response_replay.data_json)
    data["msg"] = 1337
    assert response_replay.data_json == {"msg": 1337}
    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.data_json == {"msg": 42}


def test_response_replace() -> None:
    recorded_response: _RecordedResponse = {
        "status_code": 200,
        "reason": "OK",
        "headers": {"content-type": "application/json"},
        "body": [b'{"msg":', b" 4", b"2}"],
    }
    response_replay = HTTPResponseReplay(recorded_response)
    assert response_replay.status_code == 200
    assert response_replay.reason == "OK"
    assert response_replay.headers == HTTPHeaderDict(
        {"content-type": "application/json"}
    )
    assert response_replay.data_json == {"msg": 42}

    response_replay2 = response_replay.replace(status_code=500)
    assert response_replay2.status_code == 500
    assert response_replay2.reason == "OK"
    assert response_replay2.headers == HTTPHeaderDict(
        {"content-type": "application/json"}
    )
    assert response_replay2.data_json == {"msg": 42}

    response_replay2 = response_replay.replace(reason="ALLGOOD")
    assert response_replay2.status_code == 200
    assert response_replay2.reason == "ALLGOOD"
    assert response_replay2.headers == HTTPHeaderDict(
        {"content-type": "application/json"}
    )
    assert response_replay2.data_json == {"msg": 42}

    response_replay2 = response_replay.replace(headers={"foo": "bar"})
    assert response_replay2.status_code == 200
    assert response_replay2.reason == "OK"
    assert response_replay2.headers == HTTPHeaderDict({"foo": "bar"})
    assert response_replay2.data_json == {"msg": 42}

    response_replay2 = response_replay.replace(body=b"42")
    assert response_replay2.status_code == 200
    assert response_replay2.reason == "OK"
    assert response_replay2.headers == HTTPHeaderDict(
        {"content-type": "application/json"}
    )
    assert response_replay2.data_json == 42

    response_replay2 = response_replay.replace(body=[b"13", b"37"])
    assert response_replay2.status_code == 200
    assert response_replay2.reason == "OK"
    assert response_replay2.headers == HTTPHeaderDict(
        {"content-type": "application/json"}
    )
    assert response_replay2.data_json == 1337


def test_replay_existing() -> None:
    engine = HTTPEngineReplay([REPLAY_PATH / "base"])
    request = HTTPRequest(
        method="GET",
        url="https://example.com/",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    for response in [engine(request), engine(replace(request, stream_response=True))]:
        assert response.status_code == 200
        assert response.reason == "OK"
        assert response.headers == HTTPHeaderDict({"server": "nginx"})
        assert response.data_json == {"msg": 42}

    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'HEAD', "
            "'url': 'https://example.com/', "
            "'headers': {}, "
            "'body': b''}"
        ),
    ):
        engine(replace(request, method="HEAD"))

    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'GET', "
            "'url': 'https://example.com/foobar', "
            "'headers': {}, "
            "'body': b''}"
        ),
    ):
        engine(replace(request, url="https://example.com/foobar"))

    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'GET', "
            "'url': 'https://example.com/', "
            "'headers': {'foo': 'bar'}, "
            "'body': b''}"
        ),
    ):
        engine(replace(request, headers=HTTPHeaderDict({"foo": "bar"})))

    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'GET', "
            "'url': 'https://example.com/', "
            "'headers': {}, "
            "'body': b'foobar'}"
        ),
    ):
        engine(replace(request, body=b"foobar"))


@pytest.mark.parametrize("url_end", ["wrong_ext", "in_subfolder"])
def test_replay_miss(url_end: str) -> None:
    engine = HTTPEngineReplay([REPLAY_PATH / "base"])
    request = HTTPRequest(
        method="GET",
        url=f"https://example.com/{url_end}",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )
    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'GET', "
            f"'url': 'https://example.com/{url_end}', "
            "'headers': {}, "
            "'body': b''}"
        ),
    ):
        engine(request)


def test_replay_multiple_paths() -> None:
    engine = HTTPEngineReplay([REPLAY_PATH / "base", REPLAY_PATH / "extra"])
    request = HTTPRequest(
        method="GET",
        url="https://example.com/",
        headers=HTTPHeaderDict(),
        body=b"",
        stream_response=False,
    )

    # only in base
    response = engine(request)
    assert response.data_json == {"msg": 42}

    # only in extra
    response = engine(replace(request, url="https://example.com/bar"))
    assert response.data_json == {"msg": 13}

    # in both base and extra: keep extra
    response = engine(replace(request, url="https://example.com/foo"))
    assert response.data_json == {"msg": 37}

    # in none
    with pytest.raises(
        ValueError,
        match=re.escape(
            "No response have been recorded for request: {"
            "'method': 'GET', "
            "'url': 'https://example.com/miss', "
            "'headers': {}, "
            "'body': b''}"
        ),
    ):
        engine(replace(request, url="https://example.com/miss"))


def test_replay_encoding() -> None:
    data = bytes(range(256))
    engine = HTTPEngineReplay([REPLAY_PATH / "base"])
    request = HTTPRequest(
        method="POST",
        url="https://example.com/encoding",
        headers=HTTPHeaderDict(),
        body=data,
        stream_response=False,
    )
    response = engine(request)
    assert response.data_bytes == data[::-1]


def test_replay_with_modifiers() -> None:
    def replay_request_modifier(request: HTTPRequest) -> HTTPRequest:
        assert request.url == "https://example.com/auth"
        request.url = "https://example.com/foo"
        del request.headers["authorization"]
        return request

    def replay_response_modifier(response: HTTPResponseReplay) -> HTTPResponseReplay:
        return response.replace(body=b'{"msg": 1337}')

    engine = HTTPEngineReplay(
        [REPLAY_PATH / "base"],
        replay_request_modifier=replay_request_modifier,
        replay_response_modifier=replay_response_modifier,
    )
    request = HTTPRequest(
        method="GET",
        url="https://example.com/auth",
        headers=HTTPHeaderDict({"Authorization": "Basic dXNlcjpwYXNzd29yZA=="}),
        body=b"",
        stream_response=False,
    )
    response = engine(request)
    assert response.status_code == 200
    assert response.reason == "OK"
    assert response.headers == HTTPHeaderDict({"server": "nginx"})
    assert response.data_json == {"msg": 1337}



from typing import Dict, List

# not far from string-escape encoding in Python 2
# some differences include:
#  - encoding of single-quotes
#  - decoding of "invalid" escaped-sequences like `\z`

# see also: https://github.com/python/cpython/issues/74773

_dumps_table: Dict[int, str] = {}
_loads_table: Dict[bytes, int] = {}
for _i in range(256):
    _value = f"\\x{_i:02x}"
    _loads_table[_value.encode("ascii")] = _i
    if ord(" ") <= _i <= ord("~") and _i != ord("\\"):
        _value = chr(_i)
        _loads_table[_value.encode("ascii")] = _i
    _dumps_table[_i] = _value
for _i, _value in [
    (0, "\\0"),
    (9, "\\t"),
    (10, "\\n"),
    (13, "\\r"),
    (92, "\\\\"),
]:
    _dumps_table[_i] = _value
    _loads_table[_value.encode("ascii")] = _i
del _i, _value


def stringescape_loads(data: str) -> bytes:
    try:
        remaining = memoryview(data.encode("ascii"))
    except UnicodeEncodeError as ex:
        raise ValueError("Invalid data to load (not ascii)") from ex
    out: List[int] = []
    while remaining:
        for i in range(1, 5):
            part = remaining[:i].tobytes()
            try:
                out.append(_loads_table[part])
            except KeyError:
                pass
            else:
                remaining = remaining[i:]
                break
        else:
            raise ValueError(f"Invalid data to load ({part.decode('ascii')!r})")
    return bytes(out)


def stringescape_dumps(data: bytes) -> str:
    return "".join(_dumps_table[i] for i in data)



from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from inspect import BoundArguments, signature
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar, Union
import warnings

from tenacity import (
    RetryCallState,
    Retrying,
    stop_after_attempt,
    wait_exponential_jitter,
)

from sdkite import Adapter, AdapterSpec
from sdkite.http.engine_requests import HTTPEngineRequests
from sdkite.http.exceptions import HTTPStatusCodeError
from sdkite.http.model import (
    HTTPBodyEncoding,
    HTTPHeaderDict,
    HTTPRequest,
    HTTPRequestAttemptInfo,
    HTTPResponse,
)
from sdkite.http.utils import build_status_code_check, encode_request_body, urlsjoin
from sdkite.utils import last_not_none, zip_reverse

if sys.version_info < (3, 8):  # pragma: no cover
    from typing_extensions import Literal, Protocol
else:  # pragma: no cover
    from typing import Literal, Protocol

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Callable, Iterable, Mapping
else:  # pragma: no cover
    from collections.abc import Callable, Iterable, Mapping

if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import ParamSpec
else:  # pragma: no cover
    from typing import ParamSpec


P = ParamSpec("P")
T = TypeVar("T")

HTTPAdapterSendRequest = Callable[[HTTPRequest], HTTPResponse]


_DEFAULT_RETRY_NB_ATTEMPTS = 3
_DEFAULT_WAIT_INITIAL = 1.0
_DEFAULT_WAIT_MAX = 60.0
_DEFAULT_WAIT_JITTER = 1.0


class _HTTPAdapterRequestWithoutMethodReturn(Protocol):
    def __call__(
        self,
        url: Optional[str] = None,
        *,
        body: object = None,
        body_encoding: HTTPBodyEncoding = HTTPBodyEncoding.AUTO,
        headers: Optional[Mapping[str, str]] = None,
        stream_response: bool = False,
        expected_status_codes: Union[int, str, Iterable[Union[int, str]]] = 200,
    ) -> HTTPResponse:
        ...


class _HTTPAdapterRequestWithoutMethod:
    name: str

    def __set_name__(self, klass: Any, name: str) -> None:
        self.name = name

    def __get__(
        self, instance: "HTTPAdapter", klass: Any
    ) -> _HTTPAdapterRequestWithoutMethodReturn:
        return partial(instance.request, self.name)


@dataclass
class _BeforeSleep:
    retry_callback: Optional[Callable[[HTTPRequestAttemptInfo], None]]
    initial_request: HTTPRequest

    def __call__(self, retry_call_state: RetryCallState) -> None:
        if self.retry_callback is None:
            return
        exception: BaseException = (
            retry_call_state.outcome.exception()  # type: ignore[union-attr, assignment]
        )
        seconds_since_start: float = (
            retry_call_state.seconds_since_start  # type: ignore[assignment]
        )
        self.retry_callback(
            HTTPRequestAttemptInfo(
                attempt_number=retry_call_state.attempt_number,
                initial_request=self.initial_request,
                exception=exception,
                seconds_since_start=seconds_since_start,
            )
        )


class HTTPAdapter(Adapter):
    url: Optional[str]
    headers: HTTPHeaderDict

    retry_nb_attempts: Optional[int]
    retry_callback: Optional[Callable[[HTTPRequestAttemptInfo], None]]
    retry_wait_initial: Optional[float]
    retry_wait_max: Optional[float]
    retry_wait_jitter: Optional[float]

    request_interceptor: Dict[str, int]
    response_interceptor: Dict[str, int]

    def __init__(self, send_request: Callable[[HTTPRequest], HTTPResponse]) -> None:
        self._send_request = send_request

    get = _HTTPAdapterRequestWithoutMethod()
    options = _HTTPAdapterRequestWithoutMethod()
    head = _HTTPAdapterRequestWithoutMethod()
    post = _HTTPAdapterRequestWithoutMethod()
    put = _HTTPAdapterRequestWithoutMethod()
    patch = _HTTPAdapterRequestWithoutMethod()
    delete = _HTTPAdapterRequestWithoutMethod()

    def request(
        self,
        method: str,
        url: Optional[str] = None,
        *,
        body: object = None,
        body_encoding: HTTPBodyEncoding = HTTPBodyEncoding.AUTO,
        headers: Optional[Mapping[str, str]] = None,
        stream_response: bool = False,
        expected_status_codes: Union[int, str, Iterable[Union[int, str]]] = 200,
        retry_nb_attempts: Optional[int] = None,
        retry_callback: Optional[Callable[[HTTPRequestAttemptInfo], None]] = None,
        retry_wait_initial: Optional[float] = None,
        retry_wait_max: Optional[float] = None,
        retry_wait_jitter: Optional[float] = None,
    ) -> HTTPResponse:
        check_status_code = build_status_code_check(expected_status_codes)

        #
        # create request
        #

        # method
        method = method.upper()

        # url
        url = urlsjoin(self._from_adapter_hierarchy("url", url))
        if url is None:
            raise ValueError("No URL provided")

        # headers
        _headers = headers
        headers = HTTPHeaderDict()
        for headers_part in self._from_adapter_hierarchy("headers", _headers):
            if headers_part:
                headers.update(headers_part)
        del _headers

        # body
        body, content_type = encode_request_body(body, body_encoding)
        if content_type:
            if "content-type" in headers:
                warnings.warn(
                    "The 'content-type' header is being overridden"
                    f" due to request body encoding {body_encoding}"
                    f" (from '{headers['content-type']}' to '{content_type}')",
                    UserWarning,
                    stacklevel=1,
                )
            headers["content-type"] = content_type

        # create request object
        initial_request = HTTPRequest(
            method=method,
            url=url,
            headers=headers,
            body=body,
            stream_response=stream_response,
        )

        #
        # send request
        #

        # get values from parent adapters if None, or use default
        retry_nb_attempts = last_not_none(
            self._from_adapter_hierarchy("retry_nb_attempts", retry_nb_attempts),
            _DEFAULT_RETRY_NB_ATTEMPTS,
        )
        retry_callback = last_not_none(
            self._from_adapter_hierarchy("retry_callback", retry_callback)
        )
        retry_wait_initial = last_not_none(
            self._from_adapter_hierarchy("retry_wait_initial", retry_wait_initial),
            _DEFAULT_WAIT_INITIAL,
        )
        retry_wait_max = last_not_none(
            self._from_adapter_hierarchy("retry_wait_max", retry_wait_max),
            _DEFAULT_WAIT_MAX,
        )
        retry_wait_jitter = last_not_none(
            self._from_adapter_hierarchy("retry_wait_jitter", retry_wait_jitter),
            _DEFAULT_WAIT_JITTER,
        )

        before_sleep = _BeforeSleep(retry_callback, initial_request)

        for attempt in Retrying(
            stop=stop_after_attempt(retry_nb_attempts),
            wait=wait_exponential_jitter(
                initial=retry_wait_initial,
                max=retry_wait_max,
                jitter=retry_wait_jitter,
            ),
            before_sleep=before_sleep,
            reraise=True,
        ):
            request = deepcopy(initial_request)
            with attempt:
                # request interceptors
                for interceptor in self._get_interceptors("request_interceptor"):
                    request = interceptor(request, self)

                # send request
                response = self._send_request(request)

                # response interceptors
                for interceptor in self._get_interceptors("response_interceptor"):
                    response = interceptor(response, self)

                # status code check
                if not check_status_code(response.status_code):
                    raise HTTPStatusCodeError(
                        status_code=response.status_code,
                        request=initial_request,
                        response=response,
                    )

        response._set_context(  # pylint: disable=protected-access)  # noqa: SLF001
            initial_request
        )
        return response

    def _get_interceptors(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
    ) -> List[Callable[[T, "HTTPAdapter"], T]]:
        seen_names: Set[str] = set()
        interceptors: List[Tuple[int, Callable[[T, HTTPAdapter], T]]] = []
        for client, adapter in zip_reverse(self._clients, self._adapters):
            for name, order in getattr(adapter, kind).items():
                if name not in seen_names:
                    interceptors.append((order, getattr(client, name)))
                    seen_names.add(name)
        interceptors.sort(key=lambda item: (item[0], str(item[1])))
        return [interceptor for _, interceptor in interceptors]


class HTTPAdapterSpec(AdapterSpec[HTTPAdapter]):
    _engine_callable: Callable[..., HTTPAdapterSendRequest]
    _engine_arguments: BoundArguments

    def __init__(
        self,
        url: Optional[str] = None,
        *,
        headers: Optional[Mapping[str, str]] = None,
        retry_nb_attempts: Optional[int] = None,
        retry_callback: Optional[Callable[[HTTPRequestAttemptInfo], None]] = None,
        retry_wait_initial: Optional[float] = None,
        retry_wait_max: Optional[float] = None,
        retry_wait_jitter: Optional[float] = None,
    ) -> None:
        self.url = url
        self.headers = HTTPHeaderDict(headers)

        self.retry_nb_attempts = retry_nb_attempts
        self.retry_callback = retry_callback
        self.retry_wait_initial = retry_wait_initial
        self.retry_wait_max = retry_wait_max
        self.retry_wait_jitter = retry_wait_jitter

        self.request_interceptor: Dict[str, int] = {}
        self.response_interceptor: Dict[str, int] = {}

        # defaults to engine based on 'requests'
        self.set_engine(HTTPEngineRequests)

    def set_engine(
        self,
        engine_callable: Callable[P, HTTPAdapterSendRequest],
        *engine_args: P.args,
        **engine_kwargs: P.kwargs,
    ) -> None:
        arguments = signature(engine_callable).bind(*engine_args, **engine_kwargs)
        self._engine_callable = engine_callable
        self._engine_arguments = arguments

    def _create_adapter(self) -> HTTPAdapter:
        send_request = self._engine_callable(
            *self._engine_arguments.args,
            **self._engine_arguments.kwargs,
        )
        return HTTPAdapter(send_request)

    def register_interceptor(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
        attr_name: str,
        order: int,
    ) -> None:
        interceptors: Dict[str, int] = getattr(self, kind)
        if attr_name in interceptors:
            warnings.warn(
                f"Interceptor '{attr_name}' of '{self._attr_name}' has already been registered"
                f" with order {interceptors[attr_name]}, ignoring new registration"
                f" with order {order}",
                UserWarning,
                stacklevel=1,
            )
        else:
            interceptors[attr_name] = order

    def _intercept(
        self,
        kind: Literal["request_interceptor", "response_interceptor"],
        order: int,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        def decorator(fct: Callable[P, T]) -> Callable[P, T]:
            if sys.version_info < (3, 10):  # pragma: no cover
                if isinstance(fct, (classmethod, staticmethod)):
                    name = fct.__func__.__name__
                else:
                    name = fct.__name__
            else:  # pragma: no cover
                name = fct.__name__
            self.register_interceptor(kind, name, order)
            return fct

        return decorator

    def intercept_request(
        self, order: int
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        return self._intercept("request_interceptor", order)

    def intercept_response(
        self, order: int
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        return self._intercept("response_interceptor", order)



from abc import ABC, abstractmethod
from contextlib import suppress
from dataclasses import dataclass
from enum import Enum, auto, unique
import sys
from types import TracebackType
from typing import Dict, List, Optional, Tuple, Type, Union

from sdkite.http.exceptions import HTTPContextError

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator, Mapping, MutableMapping
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator, Mapping, MutableMapping

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self


class HTTPHeaderDict(MutableMapping[str, str]):
    __slots__ = ["_contents"]

    def __init__(
        self,
        items: Union[None, Mapping[str, str], Iterable[Tuple[str, str]]] = None,
    ) -> None:
        self._contents: Dict[str, List[str]] = {}
        if isinstance(items, Mapping):
            items = items.items()
        if items:
            for key, value in items:
                self.add(key, value)

    def __repr__(self) -> str:
        as_dict = {values[0]: values[1:] for values in self._contents.values()}
        return f"HTTPHeaderDict{as_dict!r}"

    def __iter__(self) -> Iterator[str]:
        for values in self._contents.values():
            yield values[0]

    def __len__(self) -> int:
        return len(self._contents)

    def __getitem__(self, key: str) -> str:
        try:
            return ", ".join(self._contents[key.lower()][1:])
        except KeyError:
            raise KeyError(key) from None

    def __setitem__(self, key: str, value: str) -> None:
        self._contents[key.lower()] = [key, value]

    def __delitem__(self, key: str) -> None:
        del self._contents[key.lower()]

    def add(self, key: str, value: str) -> None:
        try:
            self._contents[key.lower()].append(value)
        except KeyError:
            self[key] = value


@unique
class HTTPBodyEncoding(Enum):
    AUTO = auto()
    NONE = auto()
    JSON = auto()
    URLENCODE = auto()
    MULTIPART = auto()


@dataclass
class HTTPRequest:
    method: str
    url: str
    headers: HTTPHeaderDict
    body: Union[bytes, Iterator[bytes]]
    stream_response: bool


class HTTPResponse(ABC):
    __context: Optional[HTTPRequest] = None

    @property
    @abstractmethod
    def raw(self) -> object:
        """
        The response object coming from the adapter
        """

    @property
    @abstractmethod
    def status_code(self) -> int:
        """
        The HTTP status code (RFC 2616 section 6.1.1)
        """

    @property
    @abstractmethod
    def reason(self) -> str:
        """
        The HTTP reason phrase (RFC 2616 section 6.1.1)
        """

    @property
    @abstractmethod
    def headers(self) -> HTTPHeaderDict:
        """
        The HTTP headers as a HTTPHeaderDict instance
        """

    @property
    @abstractmethod
    def data_stream(self) -> Iterator[bytes]:
        """
        The body of the response as an iterator of bytes, useful for data streaming.
        """

    @property
    @abstractmethod
    def data_bytes(self) -> bytes:
        """
        The body of the response as bytes.
        """

    @property
    @abstractmethod
    def data_str(self) -> str:
        """
        The body of the response as a str, for easier access.
        """

    @property
    @abstractmethod
    def data_json(self) -> object:
        """
        The body of the response JSON-decoded, for easier access.
        """

    def _close(self) -> None:  # noqa: B027
        pass

    def _set_context(self, context: HTTPRequest) -> None:
        self.__context = context

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        with suppress(Exception):
            self._close()
        if exc_val and self.__context:
            raise HTTPContextError.from_exception(
                exc_val,
                request=self.__context,
                response=self,
            ) from exc_val

    def __del__(self) -> None:
        with suppress(Exception):
            self._close()


@dataclass
class HTTPRequestAttemptInfo:
    attempt_number: int
    exception: BaseException
    initial_request: HTTPRequest
    seconds_since_start: float



from copy import deepcopy
from dataclasses import replace
import json
from pathlib import Path
import re
import sys
import time
from typing import Callable, Dict, List, Optional, Tuple, Union, cast

from sdkite.http._stringescape import stringescape_dumps, stringescape_loads
from sdkite.http.engine_requests import HTTPEngineRequests
from sdkite.http.model import HTTPHeaderDict, HTTPRequest, HTTPResponse
from sdkite.utils import identity

if sys.version_info < (3, 8):  # pragma: no cover
    from backports.cached_property import cached_property
    from typing_extensions import TypedDict
else:  # pragma: no cover
    from functools import cached_property
    from typing import TypedDict

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Iterable, Iterator
else:  # pragma: no cover
    from collections.abc import Iterable, Iterator


class _RecordedRequest(TypedDict):
    method: str
    url: str
    headers: Dict[str, str]
    body: bytes


class _RecordedResponse(TypedDict):
    status_code: int
    reason: str
    headers: Dict[str, str]
    body: List[bytes]


class HTTPResponseReplay(HTTPResponse):
    def __init__(self, recorded_response: _RecordedResponse) -> None:
        self.recorded_response = recorded_response

    @property
    def raw(self) -> object:
        raise ValueError("The 'raw' attribute is not available with the replay engine")

    @property
    def status_code(self) -> int:
        return self.recorded_response["status_code"]

    @property
    def reason(self) -> str:
        return self.recorded_response["reason"]

    @cached_property
    def headers(self) -> HTTPHeaderDict:
        return HTTPHeaderDict(self.recorded_response["headers"])

    @cached_property
    def data_stream(self) -> Iterator[bytes]:
        return iter(self.recorded_response["body"])

    @property
    def data_bytes(self) -> bytes:
        try:
            parts = [next(self.data_stream)]
        except StopIteration:
            raise ValueError(
                "The data_xxx attributes can be only accessed once with the replay engine"
            ) from None
        parts.extend(self.data_stream)
        return b"".join(parts)

    @property
    def data_str(self) -> str:
        return self.data_bytes.decode()

    @cached_property
    def data_json(self) -> object:
        return json.loads(self.data_bytes)

    def replace(
        self,
        *,
        status_code: Optional[int] = None,
        reason: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Union[bytes, Iterable[bytes]]] = None,
    ) -> "HTTPResponseReplay":
        recorded_response = cast(_RecordedResponse, deepcopy(self.recorded_response))
        if status_code is not None:
            recorded_response["status_code"] = status_code
        if reason is not None:
            recorded_response["reason"] = reason
        if headers is not None:
            recorded_response["headers"] = dict(headers)
        if body is not None:
            if isinstance(body, bytes):
                body = [body]
            recorded_response["body"] = list(body)
        return HTTPResponseReplay(recorded_response)


def _default_recording_compute_basename(request: HTTPRequest, _: HTTPResponse) -> str:
    return re.sub(
        "[^a-zA-Z0-9]", "_", f"{time.time():.0f} {request.method.lower()} {request.url}"
    )


class HTTPEngineReplay:
    def __init__(
        self,
        paths: Iterable[Path],
        *,
        recording: bool = False,
        replay_request_modifier: Callable[
            [HTTPRequest],
            HTTPRequest,
        ] = identity,
        replay_response_modifier: Callable[
            [HTTPResponseReplay],
            HTTPResponseReplay,
        ] = identity,
        recording_request_modifier: Callable[
            [HTTPRequest],
            HTTPRequest,
        ] = identity,
        recording_response_modifier: Callable[
            [HTTPResponseReplay],
            HTTPResponseReplay,
        ] = identity,
        recording_compute_basename: Callable[
            [HTTPRequest, HTTPResponseReplay],
            str,
        ] = _default_recording_compute_basename,
    ) -> None:
        # load all recorded items
        self.recorded: List[
            Tuple[_RecordedRequest, _RecordedResponse]
        ] = []  # sorted by priority
        for path_dir in paths:
            for path_item in path_dir.iterdir():
                if path_item.is_file() and path_item.suffix == ".json":
                    with path_item.open() as path_item_fp:
                        item = json.load(path_item_fp)
                    item["request"]["body"] = stringescape_loads(
                        item["request"]["body"]
                    )
                    item["response"]["body"] = [
                        stringescape_loads(part) for part in item["response"]["body"]
                    ]
                    self.recorded.append((item["request"], item["response"]))
            self.recording_path_dir = path_dir  # will get last item
        self.recorded.reverse()

        # set recording options
        self.recording = recording
        self.engine: Optional[Callable[[HTTPRequest], HTTPResponse]] = None

        # other attributes
        self.replay_request_modifier = replay_request_modifier
        self.replay_response_modifier = replay_response_modifier
        self.recording_request_modifier = recording_request_modifier
        self.recording_response_modifier = recording_response_modifier
        self.recording_compute_basename = recording_compute_basename

    def __call__(self, request: HTTPRequest) -> HTTPResponse:
        # exhausting body to be able to deepcopy later
        request = replace(
            request,
            body=request.body
            if isinstance(request.body, bytes)
            else b"".join(request.body),
        )

        lookup_request = self.replay_request_modifier(deepcopy(request))
        recorded_request = _RecordedRequest(
            method=lookup_request.method,
            url=lookup_request.url,
            headers=dict(lookup_request.headers),
            body=lookup_request.body
            if isinstance(lookup_request.body, bytes)
            else b"".join(lookup_request.body),
        )

        # recording mode
        if self.recording:
            # perform request
            if self.engine is None:
                self.engine = HTTPEngineRequests()

            real_request = self.recording_request_modifier(deepcopy(request))
            with self.engine(real_request) as real_response:
                received_response = HTTPResponseReplay(
                    _RecordedResponse(
                        status_code=real_response.status_code,
                        reason=real_response.reason,
                        headers=dict(real_response.headers),
                        body=list(real_response.data_stream)
                        if real_request.stream_response
                        else [real_response.data_bytes],
                    )
                )
            response = self.recording_response_modifier(received_response)

            # save recorded response
            recorded_basename = self.recording_compute_basename(request, response)
            recorded_path = self.recording_path_dir / f"{recorded_basename}.json"
            with recorded_path.open("w") as recorded_fp:
                json.dump(
                    {
                        "request": {
                            **recorded_request,
                            "body": stringescape_dumps(recorded_request["body"]),
                        },
                        "response": {
                            **response.recorded_response,
                            "body": [
                                stringescape_dumps(part)
                                for part in response.recorded_response["body"]
                            ],
                        },
                    },
                    recorded_fp,
                    indent=2,
                )
                recorded_fp.write("\n")

        # replay recorded response
        else:
            for req, resp in self.recorded:
                if req == recorded_request:
                    response = HTTPResponseReplay(resp)
                    break
            else:
                raise ValueError(
                    f"No response have been recorded for request: {recorded_request}"
                )

        return self.replay_response_modifier(response)
